//
// Created by ascadian on 7/31/18.
//

#include "gtest/gtest.h"
#include "src/Core/Intrinsics.h"
#include "src/Core/StreamCollection.h"
#include "src/Core/Serialization.h"
#include "TestUtils.h"
#include <fstream>
#include <src/Cams/TestCam/TestCam.h>
#include <boost/asio/streambuf.hpp>

namespace {

class SerializationTest : public ::testing::Test {
protected:
    SerializationTest() {

    }

    virtual ~SerializationTest() = default;
};

TEST_F(SerializationTest, serialize_intrinsics_bin) {

    {
        std::string filename("intr_serial.cereal");

        // save original intrinsics
        CamIntrinsics origIntr(1, 2, 3, 4, 5, 6, 7, 8, 9);
        {
            std::ofstream os(filename, std::ios::binary);
            cereal::PortableBinaryOutputArchive archive(os);
            archive(origIntr);
        }

        // read and compare
        {
            CamIntrinsics intr{};
            std::ifstream is(filename, std::ios::binary);
            cereal::PortableBinaryInputArchive inputArchive(is);
            inputArchive(intr);

            EXPECT_TRUE(intr == origIntr);
        }
    }
}


TEST_F(SerializationTest, serialize_intrinsics_json) {

    {
        std::string filename("intr_serial.json");

        // save original intrinsics
        CamIntrinsics origIntr(1, 2, 3, 4, 5, 6, 7, 8, 9);
        {
            std::ofstream os(filename);
            cereal::JSONOutputArchive archive(os);
            archive(origIntr);
        }

        // read and compare
        {
            CamIntrinsics intr{};
            std::ifstream is(filename);
            cereal::JSONInputArchive inputArchive(is);
            inputArchive(intr);

            EXPECT_TRUE(intr == origIntr);
        }
    }
}

TEST_F(SerializationTest, serialize_stream) {

    {
        std::string filename("stream_serial.json");

        // save original
        dyncam::TestCam cam;
        auto orig = cam.Streams.GetStreamFirst<ImageStream>(StreamTypes::COLOR);
        {
            std::ofstream os(filename);
            cereal::JSONOutputArchive archive(os);
            archive(orig);
            std::cout << orig->ToString() << std::endl;
        }

        // read and compare
        {
            auto read = std::make_shared<ImageStream>();
            std::ifstream is(filename);
            cereal::JSONInputArchive inputArchive(is);
            inputArchive(read);

            //EXPECT_TRUE(intr == origIntr);

            std::cout << read->ToString() << std::endl;
        }
    }
}

TEST_F(SerializationTest, serialize_streamcollection) {

    {
        std::string filename("streamcoll_serial.json");

        // save original
        dyncam::TestCam cam;
        auto orig = cam.Streams;
        {
            std::ofstream os(filename);
            cereal::JSONOutputArchive archive(os);
            archive(orig);
        }

        // read and compare
        {
            StreamCollection read{};
            std::ifstream is(filename);
            cereal::JSONInputArchive inputArchive(is);
            inputArchive(read);

            std::cout << orig << std::endl;
            std::cout << read << std::endl;

            //EXPECT_TRUE(orig == read);
        }
    }
}

TEST_F(SerializationTest, serialize_mat) {

    {
        std::string filename("mat_serial.json");
        std::string filename2("mat2_serial.json");
        std::string filenameBin("mat_serial.bin");

        // save original
        cv::Mat info = cv::Mat::eye(3,3, CV_32FC1);

        {
            std::ofstream os(filename);
            cereal::JSONOutputArchive archive(os);
            archive(info);
        }
        cv::Mat info2 = cv::Mat::eye(3,3, CV_8UC4);
        {
            std::ofstream os(filename2);
            cereal::JSONOutputArchive archive(os);
            archive(info2);
        }
        cv::Mat info3 = cv::Mat::eye(3,3, CV_8UC4);
        {
            std::ofstream os(filenameBin, std::ios::binary);
            cereal::PortableBinaryOutputArchive archive(os);
            archive(info2);
        }

        // read and compare
        {
            cv::Mat read{};
            std::ifstream is(filename);
            cereal::JSONInputArchive inputArchive(is);
            inputArchive(read);

            auto diff = TestUtils::compareMats(info, read);
            EXPECT_TRUE(diff == 0) << "Images difference unexpected, diff: " << cv::countNonZero(diff);
        }
        {
            cv::Mat read{};
            std::ifstream is(filename2);
            cereal::JSONInputArchive inputArchive(is);
            inputArchive(read);

            auto diff = TestUtils::compareMats(info2, read);
            EXPECT_TRUE(diff == 0) << "Images difference unexpected, diff: " << cv::countNonZero(diff);
        }
        {
            cv::Mat read{};
            std::ifstream is(filenameBin, std::ios::binary);
            cereal::PortableBinaryInputArchive inputArchive(is);
            inputArchive(read);
            auto diff = TestUtils::compareMats(info3, read);
            EXPECT_TRUE(diff == 0) << "Images difference unexpected, diff: " << cv::countNonZero(diff);
        }
    }
}

TEST_F(SerializationTest, serialize_to_streambuf) {

    {
        boost::asio::streambuf buff;
        // save original
        cv::Mat info = cv::Mat::eye(3000,3000, CV_32FC1);
        {
            std::ostream os(&buff);
            cereal::PortableBinaryOutputArchive archive(os);
            archive(info);
            size_t bodySize = buff.size();
            std::cout << "Serialized: " << bodySize << std::endl;
        }
    }
}


}